/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.getModelElementAncestor;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getModelElementLevel;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.fortiss.tooling.kernel.model.INamedElement;

/**
 * Comparator that sorts rows according to hierarchy level. Elements within one hierarchy level
 * are sorted by name.
 * 
 * @author eder, diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 34087E50E0586FFC5796616F703D9C08
 */
public class HierarchicalNameComparator implements java.util.Comparator<EObject> {

	/**
	 * Compare two {@link Object}s by
	 * <ul>
	 * <li>name (if the are {@link INamedElement}s)</li>
	 * <li>Id (if they are {@link IIdLabeled} elements)</li>
	 * <li>Hash code (otherwise)</li>
	 * </ul>
	 */
	public static int compareByNameIdHashCode(Object o1, Object o2) {
		// Compare elements by name if they are INamedElements
		if(o1 instanceof INamedElement && o2 instanceof INamedElement) {
			// Elements are at the same level -> compare by name
			String name1 = ((INamedElement)o1).getName();
			String name2 = ((INamedElement)o2).getName();
			if(name1 == null) {
				name1 = "";
			}
			if(name2 == null) {
				name2 = "";
			}
			int rval = name1.compareTo(name2);
			if(rval != 0) {
				return rval;
			}
		}

		if(o1 instanceof IIdLabeled && o2 instanceof IIdLabeled) {
			return Integer.compare(((IIdLabeled)o1).getId(), ((IIdLabeled)o2).getId());
		}

		return Integer.compare(o1.hashCode(), o2.hashCode());
	}

	/** {@inheritDoc} */
	@Override
	public int compare(EObject modelElement1, EObject modelElement2) {

		int modelElement1Level = getModelElementLevel(modelElement1);
		int modelElement2Level = getModelElementLevel(modelElement2);

		EObject e1 = modelElement1;
		EObject e2 = modelElement2;

		if(modelElement1Level < modelElement2Level) {
			// Element1 is higher than element2: Get ancestor of element2 that is at the
			// level as the element1.
			int levelsUp = modelElement2Level - modelElement1Level;
			EObject ancestor = getModelElementAncestor(modelElement2, levelsUp);

			if(modelElement1 == ancestor) {
				// Element1 is offspring of element2
				return -1;
			}

			// Further comparison starts on elements at the same level
			e2 = ancestor;
		} else if(modelElement1Level > modelElement2Level) {
			// Element2 is higher than element1: Get ancestor of element1 that is at the
			// level as the element2.
			int levelsUp = modelElement1Level - modelElement2Level;
			EObject ancestor = getModelElementAncestor(modelElement1, levelsUp);

			if(modelElement2 == ancestor) {
				// Element2 is offspring of element1
				return 1;
			}

			// Further comparison starts on elements at the same level
			e1 = ancestor;
		}
		// Find direct children of common root element of element1 and element2...
		do {
			if(e1.eContainer() == e2.eContainer()) {
				break;
			}
			e1 = e1.eContainer();
			e2 = e2.eContainer();
		} while(true);

		// ... and compare them
		return compareByNameIdHashCode(e1, e2);
	}
}
