/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view.generic;

import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.valueprovider.AnnotationInstSpec;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;

/**
 * Data required to identify a column displaying a particular {@link IAnnotatedSpecification} in
 * a column of the {@link GenericAnnotationView}. Used to sort columns (see
 * {@link #compareTo(ColumnHandle)}).
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: D2398D01DFE1F8E87DAB7820C808E928
 */
public class ColumnHandle<T extends IAnnotatedSpecification> extends AnnotationInstSpec<T>
		implements Comparable<ColumnHandle<T>> {
	/**
	 * Character used to separate the annotation name from the instance name for multi-instance
	 * annotations.
	 */
	private final static String MULTIINSTANCE_NAME_SEPARATOR = ":";

	/** Flag if this {@link ColumnHandle} represents a {@link CreateAnnotationInstanceColumn}. */
	private boolean isCreateColumn;

	/** Constructs a new {@link ColumnHandle}. */
	@SuppressWarnings("unchecked")
	public ColumnHandle(AnnotationEntry entry, T annotatedSpecification, String instanceKey,
			boolean isCreateColumn) {
		super(entry, (Class<T>)annotatedSpecification.getClass(), annotatedSpecification,
				instanceKey);
		this.isCreateColumn = isCreateColumn;
	}

	/**
	 * Returns a flag if this {@link ColumnHandle} represents a
	 * {@link CreateAnnotationInstanceColumn}.
	 */
	public boolean isCreateColumn() {
		return isCreateColumn;
	}

	/**
	 * Returns the name of a column for a given {@link IAnnotatedSpecification} represented by
	 * the given {@link ColumnHandle}.
	 */
	public String getColumnName() {
		// Determine column name
		String specName =
				getEntry().getSpecificationAnnotationName(getAnnotatedSpecification().getClass());

		if(getInstanceKey() != null) {
			if(!specName.isEmpty()) {
				specName += MULTIINSTANCE_NAME_SEPARATOR + " ";
			}
			specName += getInstanceKey();
		} else if(specName == null || specName.isEmpty()) {
			specName = "<Unnamed Annotation>";
		}
		return specName;
	}

	/** {@inheritDoc} */
	@Override
	public int compareTo(ColumnHandle<T> other) {
		String columnName = getColumnName();
		String annotationName;

		int columnNameSeparatorIndex = columnName.indexOf(MULTIINSTANCE_NAME_SEPARATOR);
		if(columnNameSeparatorIndex != -1) {
			annotationName = columnName.substring(0, columnNameSeparatorIndex);
		} else {
			annotationName = columnName;
		}

		String otherColumnName = other.getColumnName();
		String otherAnnotationName;
		int otherColumnNameSeparatorIndex = otherColumnName.indexOf(MULTIINSTANCE_NAME_SEPARATOR);
		if(otherColumnNameSeparatorIndex != -1) {
			otherAnnotationName = otherColumnName.substring(0, otherColumnNameSeparatorIndex);
		} else {
			otherAnnotationName = otherColumnName;
		}

		// Place "create" columns (label: "<specification name>") after their instances
		// (which are labeled "<specification name>: <instance name>").
		if(annotationName.equals(otherAnnotationName)) {
			// Ensure uniqueness of "create" columns"
			if(isCreateColumn && other.isCreateColumn) {
				return 0;
			}
			if(isCreateColumn) {
				return 1;
			}
			if(other.isCreateColumn) {
				return -1;
			}
		}

		// Sort alphabetically by specification column label
		return columnName.compareTo(otherColumnName);
	}
}
