/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2015 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout.auto;

import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTOR_ANGLE;
import static org.fortiss.tooling.base.utils.AngleUtils.getAngle;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.resource.ImageDescriptor;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.ui.ToolingBaseUIActivator;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.ui.extension.IContextMenuContributor;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;

/**
 * Context menu entry to automatic layout {@link IHierarchicElement}-based models.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 76FF1ADFC034DE69815575B04BA534F0
 */
public class AutoLayoutMenu implements IContextMenuContributor {

	/** String appearing in the context menu. */
	private static String MENU_NAME = "Automatic layout";

	/** Declares where is located the icon to use for the menu contribution */
	private static String MENU_ICON = "icons/autolayout.png";

	/** Returns the icon that is visible in the context menu for this entry. */
	protected ImageDescriptor getActionIcon() {
		return ToolingBaseUIActivator.getImageDescriptor(MENU_ICON);
	}

	/** {@link Action} to execute automatic layout. */
	protected class LaunchAutoLayoutAction extends Action {

		/** {@link IHierarchicElement} selected by user */
		private final IHierarchicElement element;

		/** Constructor. */
		public LaunchAutoLayoutAction(IHierarchicElement recGraph) {
			super(MENU_NAME, getActionIcon());
			this.element = recGraph;
		}

		/** {@inheritDoc} */
		@Override
		public void run() {
			IAutoLayouter layouter = new KielerAutoLayouter();
			layouter.performAutoLayout(element);
		}
	}

	/**
	 * Predicate if the given {@link IHierarchicElement} contains an {@link IConnector} whose
	 * position is specified using an angle, and that is connected to at least one incoming or
	 * outgoing {@link IConnection}.
	 */
	private boolean hasConnectedConnectorWithAngle(IHierarchicElement element) {
		for(IConnector connector : element.getConnectors()) {
			if(getAngle((ILayoutedModelElement)connector, CONNECTOR_ANGLE) != null) {
				return !connector.getIncoming().isEmpty() || !connector.getOutgoing().isEmpty();
			}
		}

		return false;
	}

	/** {@inheritDoc} */
	@Override
	public List<IContributionItem> getContributedItems(EObject selection,
			ContextMenuContextProvider contextProvider) {

		if(!(selection instanceof IHierarchicElement)) {
			return Collections.emptyList();
		}

		IHierarchicElement element = (IHierarchicElement)selection;
		// Skip single top-level model element (e.g., AF3 component architecture)
		if(element instanceof IProjectRootElement && element.getContainedElements().size() == 1) {
			// There is exactly one child element
			element = element.getContainedElements().get(0);
		}

		// Check if model is empty
		if(pickInstanceOf(ILayoutedModelElement.class, element.getContainedElements()).isEmpty()) {
			return Collections.emptyList();
		}

		// Enable automatic layouter only for diagrams without rounded edges, which is the case if
		// none of the positions of the connected connectors is specified using an angle (i.e., in
		// particular also if the model does not contain any connectors, or none of the connectors
		// is connected to a connection).
		if(hasConnectedConnectorWithAngle(element)) {
			return Collections.emptyList();
		}
		for(IHierarchicElement childElement : element.getContainedElements()) {
			if(hasConnectedConnectorWithAngle(childElement)) {
				return Collections.emptyList();
			}
		}

		List<IContributionItem> contributionItems = new ArrayList<IContributionItem>();
		contributionItems.add(new ActionContributionItem(new LaunchAutoLayoutAction(element)));
		return contributionItems;
	}

	/** {@inheritDoc} */
	@Override
	public String getMenuSectionID() {
		return null;
	}
}
