/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.getActiveConstraintsTransitively;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.removeFromConstraintInstanceContainer;

import java.util.List;
import java.util.function.Function;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.fortiss.tooling.base.model.base.ConstraintBasedProcess;
import org.fortiss.tooling.base.model.element.ConstraintConfiguration;
import org.fortiss.tooling.base.model.element.IConstraintBasedProcess;
import org.fortiss.tooling.kernel.extension.IConstraint;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceStatus;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.listener.IPersistencyServiceListener;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Utility methods related to constraints.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: CBB153EAF0C3F84123593CDBAD3B9FDD
 */
public class ConstraintsBaseUIUtils {

	/** Factorized behaviour. */
	public static void doActivateAll(IConstraintInstanceContainer cstrContainer,
			IConstraintBasedProcess cbdProcess) {
		// First activate the constraints which should always be activated
		for(Class<? extends IConstraint> cstrClass : IConstraintUIService.getInstance()
				.getAlwaysActivatedConstraints()) {
			IConstraintUIService.getInstance().activate(cstrClass, cstrContainer);
		}
		// Then activate the constraints of the current objective
		ConstraintConfiguration currentObj = cbdProcess.getCurrentObjective();
		if(currentObj != null) {
			List<Class<? extends IConstraint>> activeCstrs =
					getActiveConstraintsTransitively(currentObj, null);
			for(Class<? extends IConstraint> cstrClass : activeCstrs) {
				IConstraintUIService.getInstance().activate(cstrClass, cstrContainer);
			}
		}
	}

	/**
	 * Activates all constraints which are active in the given configuration, potentially adding new
	 * constraint instances to the provided container.
	 */
	public static void activateConfiguration(ConstraintConfiguration config,
			IConstraintInstanceContainer cstrContainer) {
		for(Class<? extends IConstraint> cstr : getActiveConstraintsTransitively(config, null)) {
			IConstraintUIService.getInstance().activate(cstr, cstrContainer);
		}
	}

	/**
	 * Deactivates all constraints which are active in the given configuration, if they are not
	 * still active in the provided context.
	 */
	public static void deactivateConfiguration(ConstraintConfiguration config,
			IConstraintInstanceContainer cstrContainer, ConstraintConfiguration context) {
		List<Class<? extends IConstraint>> stillActiveConstraints =
				getActiveConstraintsTransitively(context, config);
		for(Class<? extends IConstraint> cstr : getActiveConstraintsTransitively(config, null)) {
			if(!stillActiveConstraints.contains(cstr)) {
				IConstraintUIService.getInstance().deactivate(cstr, cstrContainer);
			}
		}
	}

	/**
	 * Factorized behavior to install the various listeners necessary for constraints.
	 * This made in a generic manner so that the client only needs to give a function providing
	 * access to a {@link ConstraintBasedProcess} given a {@link ITopLevelElement}. The rest is for
	 * free.
	 */
	public static class ConstraintBasedProcessAwareActivator {
		/**
		 * Function to access a {@link ConstraintBasedProcess} given a {@link ITopLevelElement}.
		 * Project-dependent.
		 */
		private Function<ITopLevelElement, ConstraintBasedProcess> getConstraintBasedProcess;

		/** Constructor. */
		public ConstraintBasedProcessAwareActivator(
				Function<ITopLevelElement, ConstraintBasedProcess> getConstraintBasedProcess) {
			this.getConstraintBasedProcess = getConstraintBasedProcess;
		}

		/** Start the kernel. */
		public void start() {
			for(ITopLevelElement top : IPersistencyService.getInstance().getTopLevelElements()) {
				installListener(top);
				activateAllConstraints(top);
			}
			IPersistencyService.getInstance().addTopLevelElementListener(
					new ConstraintBasedDevelopmentProcessListener());
		}

		/** Go through all constraints in <code>top</code> and (re-)activates them. */
		private void activateAllConstraints(ITopLevelElement top) {
			ConstraintBasedProcess cbp = getConstraintBasedProcess.apply(top);
			if(cbp != null) {
				IConstraintInstanceContainer cic = cbp.getConstraintInstanceContainer();
				try {
					doActivateAll(cic, cbp);
					// The only case where the activation above can trigger a model change is if it
					// is discovered that constraints shall be added.
					// This should not happen in most cases: the constraint was normally active when
					// the project was left and therefore the constraint should already be present.
					// However there is a case where this is not the case: when AF3 developers added
					// a new constraint and start then AF3 for the first time; then the constraint
					// has to be added (sort of migration of the model actually). In such a case the
					// line above will trigger an exception because not run in a command. That is
					// the whole reason for the catch block thereafter.
					// Note that we do not run the activation directly in a command because we do
					// not want to mark the model as dirty for nothing.
				} catch(IllegalStateException e) {
					top.runAsCommand(() -> doActivateAll(cic, cbp));
				}
			}
		}

		/**
		 * Install the listener triggering activation of constraints for the given top level
		 * element.
		 * Returns the installed listener.
		 */
		private void installListener(ITopLevelElement top) {
			ConstraintBasedProcess cbp = getConstraintBasedProcess.apply(top);
			if(cbp == null) {
				return;
			}
			IConstraintInstanceContainer cic = cbp.getConstraintInstanceContainer();
			EContentAdapter listener = new EContentAdapter() {

				private boolean isAlreadyAddingMissingConstraints = false;

				/** {@inheritDoc} */
				@Override
				public void notifyChanged(Notification notification) {
					super.notifyChanged(notification);
					if(isTouch(notification) || isAlreadyAddingMissingConstraints) {
						return;
					}
					// Do not react to change of status.
					if(notification.getNewValue() instanceof IConstraintInstanceStatus) {
						return;
					}
					if(notification.getEventType() == Notification.REMOVE &&
							notification.getOldValue() instanceof IConstrained) {
						IConstrained constrained = (IConstrained)notification.getOldValue();
						removeFromConstraintInstanceContainer(constrained);
					}
					isAlreadyAddingMissingConstraints = true;
					IConstraintUIService.getInstance().addMissingConstraintInstances(cic);
					isAlreadyAddingMissingConstraints = false;
				}
			};
			top.getRootModelElement().eAdapters().add(listener);
		}

		/**
		 * Listener to (re)activate constraints when the project changes.
		 * Typically to install adequate constraints when a new element was added to the project.
		 */
		private class ConstraintBasedDevelopmentProcessListener implements
				IPersistencyServiceListener {

			/** {@inheritDoc} */
			@Override
			public void topLevelElementAdded(ITopLevelElement element) {
				installListener(element);
			}

			/** {@inheritDoc} */
			@Override
			public void topLevelElementContentChanged(ITopLevelElement element) {
				// Not our concern.
			}

			/** {@inheritDoc} */
			@Override
			public void topLevelElementRemoved(ITopLevelElement element) {
				// Not our concern
			}
		}

		/**
		 * @param notification
		 * @return true if <code>notification</code> is not relevant and should be skipped
		 */
		private static boolean isTouch(Notification notification) {
			// Notifications which do not notify about a model change.
			if(notification.isTouch()) {
				return true;
			}
			// Notifications initiated by the the AF3 kernel which are only GUI related.
			if(notification.getEventType() == EcoreUtils.KERNEL_EMF_EVENT_TYPE_COUNT) {
				return true;
			}
			return false;
		}
	}
}
