/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createDimension;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createPoint;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.gef.RootEditPart;
import org.eclipse.gef.editparts.ScalableFreeformRootEditPart;
import org.eclipse.gef.editparts.ScalableRootEditPart;
import org.eclipse.gef.requests.ChangeBoundsRequest;
import org.fortiss.tooling.base.layout.DefaultLayoutConstants;
import org.fortiss.tooling.base.utils.ZoomUtils;

/**
 * Utility code to support zooming.
 * 
 * @author hummel
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: FDBEA15E01470E5D318805D3653A4F3F
 */
public class ZoomUIUtils {

	/** Factor to align zoomed movements to 0.05 steps (==> avail. zoom factors). */
	private static int ZOOM_ALIGN_FACTOR = 20;

	/** Extracts the current zoom level from the given root edit part. */
	public static double determineZoom(RootEditPart rootEditPart) {
		if(rootEditPart instanceof ScalableRootEditPart) {
			return ((ScalableRootEditPart)rootEditPart).getZoomManager().getZoom();
		}
		if(rootEditPart instanceof ScalableFreeformRootEditPart) {
			return ((ScalableFreeformRootEditPart)rootEditPart).getZoomManager().getZoom();
		}
		return 1;
	}

	/**
	 * Fixes a {@link ChangeBoundsRequest} for the current zoom level. Both the
	 * move and size delta are changed.
	 */
	public static void fixChangeBoundRequest(ChangeBoundsRequest request, double zoom) {
		if(zoom <= 0 || zoom == 1) {
			return;
		}

		Integer numElem = (Integer)request.getExtendedData().get("SELECTED_ELEM_NUM");
		if(numElem == null) {
			// Do not consider elements that are part of the parent pane here, e.g. Ports.
			// They are irrelevant for the calculation of the scale factor and would cause "jumping"
			// of the selected elements.
			return;
		}

		if(numElem != 1) {
			zoom = 1 - (1 - zoom) / numElem;
			zoom = Math.floor(zoom * ZOOM_ALIGN_FACTOR) / ZOOM_ALIGN_FACTOR;
		}

		request.setMoveDelta(screenToInternalPoint(request.getMoveDelta(), zoom));
		request.setSizeDelta(screenToInternalDimension(request.getSizeDelta(), zoom));

		// Adjust to grid (snap).
		Point reqPoint = request.getMoveDelta();
		SnapToGridUtils.snapToGrid(reqPoint,
				(int)Math.round(DefaultLayoutConstants.DEFAULT_GRID_SIZE * zoom));
	}

	/**
	 * Converts a point expressed in screen coordinates, into a point in internal coordinates,
	 * taking the zoom into account.
	 */
	public static Point screenToInternalPoint(Point pos, double zoom) {
		org.fortiss.tooling.base.model.layout.Point modelPoint =
				createPoint(pos.x(), pos.y(), pos.toString());
		modelPoint = ZoomUtils.screenToInternalPoint(modelPoint, zoom);
		pos.setLocation(modelPoint.getX(), modelPoint.getY());
		return pos;
	}

	/**
	 * Converts a point expressed in internal coordinates, into a point in screen coordinates,
	 * taking the zoom into account.
	 */
	public static Point internalToScreenPoint(Point pos, double zoom) {
		org.fortiss.tooling.base.model.layout.Point modelPoint =
				createPoint(pos.x(), pos.y(), pos.toString());
		modelPoint = ZoomUtils.internalToScreenPoint(modelPoint, zoom);
		pos.setLocation(modelPoint.getX(), modelPoint.getY());
		return pos;
	}

	/**
	 * Converts a point expressed in screen coordinates, into a point in internal coordinates,
	 * taking the zoom into account.
	 */
	public static Dimension screenToInternalDimension(Dimension dim, double zoom) {
		org.fortiss.tooling.base.model.layout.Dimension modelDim =
				createDimension(dim.width(), dim.height(), dim.toString());
		modelDim = ZoomUtils.screenToInternalDimension(modelDim, zoom);
		dim.setSize(modelDim.getWidth(), modelDim.getHeight());
		return dim;
	}

	/**
	 * Converts a point expressed in internal coordinates, into a point in screen coordinates,
	 * taking the zoom into account.
	 */
	// TODO(VA) This method is not used -> remove?
	public static Dimension internalToScreenDimension(Dimension dim, double zoom) {
		org.fortiss.tooling.base.model.layout.Dimension modelDim =
				createDimension(dim.width(), dim.height(), dim.toString());
		modelDim = ZoomUtils.internalToScreenDimension(modelDim, zoom);
		dim.setSize(modelDim.getWidth(), modelDim.getHeight());
		return dim;
	}
}
