/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2012 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.graphicsGL.util;

import static org.conqat.lib.commons.filesystem.FileSystemUtils.readLinesUTF8;
import static org.fortiss.tooling.graphicsGL.util.Graphics3DModelElementFactory.color;
import static org.fortiss.tooling.graphicsGL.util.Mesh3DModelElementFactory.material;
import static org.fortiss.tooling.graphicsGL.util.Mesh3DModelElementFactory.materialGroup;
import static org.fortiss.tooling.graphicsGL.util.Mesh3DModelElementFactory.mesh;
import static org.fortiss.tooling.graphicsGL.util.Mesh3DModelElementFactory.normal;
import static org.fortiss.tooling.graphicsGL.util.Mesh3DModelElementFactory.triangleFace;
import static org.fortiss.tooling.graphicsGL.util.Mesh3DModelElementFactory.vertex;
import static org.fortiss.tooling.graphicsGL.util.Scene3DModelElementFactory.object3D;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.fortiss.tooling.graphicsGL.model.Color3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Face3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Material3D;
import org.fortiss.tooling.graphicsGL.model.mesh.MaterialGroup3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Mesh3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Normal3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Vertex3D;
import org.fortiss.tooling.graphicsGL.model.scene.Object3D;

/**
 * Utility class for loading OBJ files (i.e. Wavefront format: .obj and .mtl files).
 * These files can be produced by exporting 3D objects from Blender (http://www.blender.org).
 * Note that you must enable the export normals option in order to get the normal vectors, too.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 3916F8F842BC95EBC19A75ACFAA5C90E
 */
public class OBJLoaderUtil {
	/** Material map. */
	private Map<String, Material3D> materials = new HashMap<String, Material3D>();
	/** Vertex list of current object. */
	private List<Vertex3D> vertices = new ArrayList<Vertex3D>();
	/** Normals list of current object. */
	private List<Normal3D> normals = new ArrayList<Normal3D>();
	/** Face list of current object and current material. */
	private List<Face3D> faces = new ArrayList<Face3D>();
	/** Material group list of the current object. */
	private List<MaterialGroup3D> matGroups = new ArrayList<MaterialGroup3D>();

	/** Material of current object. */
	private Material3D material = null;
	/** Name of the current object. */
	private String objectName = null;
	/** Boolean flag for smooth shading. */
	private boolean smoothShading = false;

	/** The data for the loaded objects. */
	private List<Object3D> objectData;

	/** Constructor for file input with material file derived from this data. */
	public OBJLoaderUtil(File objFile) throws IOException {
		List<String> objData = readLinesUTF8(objFile);
		for(String line : objData) {
			if(line.startsWith("mtllib ")) { // material library reference found
				File matFile = new File(objFile.getParentFile(), line.split(" ")[1]);
				loadMaterial(readLinesUTF8(matFile));
			}
		}
		loadVerticesAndNormals(objData);
		loadObjects(objData);
	}

	/** Constructor for direct object and material content. */
	public OBJLoaderUtil(List<String> objectData, List<String> materialData) {
		loadMaterial(materialData);
		loadVerticesAndNormals(objectData);
		loadObjects(objectData);
	}

	/** Loads all vertex and normals. */
	private void loadVerticesAndNormals(List<String> objData) {
		for(String line : objData) {
			if(line.startsWith("#")) { // comment
				continue;
			} else if(line.startsWith("vn ")) {
				addNormal(line.split(" "));
			} else if(line.startsWith("v ")) {
				addVertex(line.split(" "));
			}
		}
	}

	/** Returns the list of loaded objects. */
	public List<Object3D> getLoadedObjects() {
		return objectData;
	}

	/** Loads the model objects. */
	private void loadObjects(List<String> objData) {
		objectData = new LinkedList<Object3D>();
		for(String line : objData) {
			if(line.startsWith("#")) { // comment
				continue;
			} else if(line.startsWith("f ")) {
				addFace(line.split(" "));
			} else if(line.startsWith("s ")) {
				if(line.split(" ")[1].equals("1")) {
					smoothShading = true;
				}
			} else if(line.startsWith("o ")) { // next object
				addMaterialGroup();
				addObject(objectData);
				objectName = line.split(" ")[1];
			} else if(line.startsWith("usemtl ")) { // use material
				addMaterialGroup();
				String mat = line.split(" ")[1];
				material = materials.get(mat);
				if(material == null) {
					throw new IllegalArgumentException("Unknown material used in OBJ file: " + mat);
				}
			}
		}
		addMaterialGroup();
		addObject(objectData); // last object;
	}

	/** Adds a material group and clears temporary data. */
	private void addMaterialGroup() {
		if(material != null) {
			matGroups.add(materialGroup(material, faces, smoothShading));
			material = null;
			smoothShading = false;
			faces.clear();
		}
	}

	/** Adds an object if some faces exists and clear temporary data. */
	private void addObject(List<Object3D> result) {
		if(!matGroups.isEmpty()) {
			Mesh3D mesh = mesh(matGroups);
			result.add(object3D(objectName, "", 0, 0, 0, mesh));
			objectName = null;
			material = null;
			faces.clear();
			matGroups.clear();
		}
	}

	/** Adds a face. */
	private void addFace(String[] tokens) {
		if(tokens.length != 4) {
			throw new IllegalArgumentException("OBJ file contains non-triangles.");
		}
		int[] v1 = getVertexData(tokens[1]);
		int[] v2 = getVertexData(tokens[2]);
		int[] v3 = getVertexData(tokens[3]);
		Face3D f =
				triangleFace(vertices.get(v1[0]), vertices.get(v2[0]), vertices.get(v3[0]),
						normals.get(v1[1]), normals.get(v2[1]), normals.get(v3[1]));
		faces.add(f);
	}

	/** Returns the indices in the vertex and normals list. Texture is currently ignored. */
	private int[] getVertexData(String s) {
		String[] ids = s.split("/");
		if(ids.length != 3) {
			throw new IllegalArgumentException("OBJ file does not contain normal vectors.");
		}
		return new int[] {Integer.valueOf(ids[0]) - 1, Integer.valueOf(ids[2]) - 1};
	}

	/** Adds a vertex. */
	private void addVertex(String[] tokens) {
		vertices.add(vertex(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
				Float.valueOf(tokens[3])));
	}

	/** Adds a normal. */
	private void addNormal(String[] tokens) {
		normals.add(normal(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
				Float.valueOf(tokens[3])));
	}

	/** Loads the materials from the given URL and stores them in the map. */
	public void loadMaterial(List<String> lines) {
		String matName = null;
		Color3D ambient = null;
		Color3D diffuse = null;
		Color3D specular = null;
		float specularCoeff = 0;

		for(String line : lines) {
			if(line.startsWith("#")) { // comment
				continue;
			}
			String[] tokens = line.split(" ");
			if(tokens.length > 1) {
				if("newmtl".equals(tokens[0])) {
					if(matName != null) {
						Material3D newMat = material(ambient, diffuse, specular, specularCoeff);
						materials.put(matName, newMat);
						ambient = color(0.0f, 0.0f, 0.0f);
						diffuse = color(0.0f, 0.0f, 0.0f);
						specular = color(0.0f, 0.0f, 0.0f);
						specularCoeff = 0;
					}
					matName = tokens[1];
				} else if("Ka".equals(tokens[0])) {
					ambient =
							color(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
									Float.valueOf(tokens[3]));
				} else if("Kd".equals(tokens[0])) {
					diffuse =
							color(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
									Float.valueOf(tokens[3]));
				} else if("Ks".equals(tokens[0])) {
					specular =
							color(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
									Float.valueOf(tokens[3]));
				} else if("Ns".equals(tokens[0])) {
					specularCoeff = Float.valueOf(tokens[1]);
				}
			}
		}

		if(matName != null) {
			Material3D newMat = material(ambient, diffuse, specular, specularCoeff);
			materials.put(matName, newMat);
		}
	}
}
