/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.EventObject;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.common.command.AbstractCommand;
import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.internal.storage.eclipse.AutoUndoCommandStack;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * The dummy top-level element is used to load models into the {@link IPersistencyService} without
 * an actual underlying storage. This is primarily intended to load models for unit testing.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 53DB75B1F9A5FF42276841067F065CB1
 */
final class DummyTopLevelElement implements ITopLevelElement, CommandStackListener {

	/** Stores the root element. */
	private final EObject root;

	/** The editing domain used for this model. */
	private final TransactionalEditingDomain editingDomain;

	/** The transactional command stack. */
	private final AutoUndoCommandStack transactionalCommandStack;

	/** Stores the command stack listener for this context. */
	private final List<CommandStackListener> listeners = new LinkedList<CommandStackListener>();

	/** Constructor. */
	public DummyTopLevelElement(EObject root) {
		Assert.isNotNull(root);
		this.root = root;

		editingDomain = TransactionalEditingDomain.Factory.INSTANCE.createEditingDomain();

		Resource res =
				editingDomain.getResourceSet().createResource(URI.createURI("dummy://dummy"));
		transactionalCommandStack = new AutoUndoCommandStack(editingDomain);
		transactionalCommandStack.addCommandStackListener(this);
		runAsNonDirtyingCommand(() -> res.getContents().add(root));
	}

	/** {@inheritDoc} */
	@Override
	public EObject getRootModelElement() {
		return root;
	}

	/** {@inheritDoc} */
	@Override
	public void runAsCommand(final Runnable runnable) {
		transactionalCommandStack.execute(new AbstractCommand() {

			@Override
			public boolean canExecute() {
				return true;
			}

			@Override
			public void execute() {
				runnable.run();
			}

			@Override
			public void redo() {
				// we do nothing here, as redo is handled by our command stack.
			}
		});
	}

	/** {@inheritDoc} */
	@Override
	public void runAsNonDirtyingCommand(final Runnable runnable) {
		try {
			transactionalCommandStack.executeNonDirtyingNonUndoing(new AbstractCommand() {
				@Override
				public void execute() {
					runnable.run();
				}

				@Override
				public boolean canExecute() {
					return true;
				}

				@Override
				public void redo() {
					// we do nothing here, as redo is handled by our command stack.
				}

				/** {@inheritDoc} */
				@Override
				public boolean canUndo() {
					return false;
				}
			});
		} catch(Exception e) {
			e.printStackTrace();
			error(ToolingKernelActivator.getDefault(),
					"Problem with synchronizing library and model!", e);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void addCommandStackListener(CommandStackListener listener) {
		listeners.add(listener);
	}

	/** {@inheritDoc} */
	@Override
	public void removeCommandStackListener(CommandStackListener listener) {
		listeners.remove(listener);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canUndo() {
		return transactionalCommandStack.canUndo();
	}

	/** {@inheritDoc} */
	@Override
	public boolean canRedo() {
		return transactionalCommandStack.canRedo();
	}

	/** {@inheritDoc} */
	@Override
	public void undo() {
		transactionalCommandStack.undo();
	}

	/** {@inheritDoc} */
	@Override
	public void redo() {
		transactionalCommandStack.redo();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isDirty() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public String getSaveableName() {
		return "Dummy";
	}

	/** {@inheritDoc} */
	@Override
	public void doSave(IProgressMonitor monitor) {
		// ignore
	}

	/** {@inheritDoc} */
	@Override
	public void doSaveAndResetLastChangeWasIntended(IProgressMonitor monitor) {
		// ignore
	}

	/** {@inheritDoc} */
	@Override
	public String getId(EObject modelElement) {
		if(modelElement instanceof IIdLabeled) {
			return String.valueOf(((IIdLabeled)modelElement).getId());
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void prepareIDs(EObject modelElement) {
		if(!(modelElement instanceof IIdLabeled))
			return;

		int id = Integer.MIN_VALUE;
		for(TreeIterator<EObject> iter = getRootModelElement().eAllContents(); iter.hasNext();) {
			EObject o = iter.next();
			if(o instanceof IIdLabeled) {
				id = Math.max(id, ((IIdLabeled)o).getId());
			}
		}

		((IIdLabeled)modelElement).setId(id + 1);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canDelete() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean delete() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public ResourceSet getResourceSet() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void commandStackChanged(EventObject event) {
		EventObject eo = new EventObject(this);
		for(CommandStackListener l : listeners) {
			l.commandStackChanged(eo);
		}
	}
}
