/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.util.Collections.emptyList;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.conqat.lib.commons.reflect.ReflectionUtils;
import org.fortiss.tooling.kernel.extension.ITransformationProvider;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;
import org.fortiss.tooling.kernel.extension.data.TransformationProviderChain;
import org.fortiss.tooling.kernel.extension.exception.TransformationFailedException;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.introspection.items.TransformationServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.ITransformationService;
import org.fortiss.tooling.kernel.service.base.ObjectAware2ServiceBase;

/**
 * This class implements the {@link ITransformationService} interface.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 31218DC40D5050DC5F71ED998306FE84
 */
public class TransformationService extends ObjectAware2ServiceBase<ITransformationProvider>
		implements ITransformationService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final TransformationService INSTANCE = new TransformationService();

	/** Returns singleton instance of the service. */
	public static TransformationService getInstance() {
		return INSTANCE;
	}

	/** The connector extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.transformationProvider";

	/** The connector configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "transformationProvider";

	/** The connector attribute name. */
	private static final String ATTRIBUTE_NAME = "transformationProvider";

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service manages registration and execution of transformations." +
				"\nThis service can be asked to provide a transformation or a chain of transformations, which " +
				"\ntransform an object of arbitrary class A into an object of another arbitrary class B." +
				"\nTherefore transformations are not restricted to EObject instances, but extend to all Java classes." +
				"\nWhen a specific transformation is required the service can also provide a chain of transformations" +
				"\nbased on the providers available to it." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** Registers the given transformation with the service. */
	@Override
	public void registerTransformationProvider(ITransformationProvider provider,
			Class<?> sourceClass, Class<?> targetClass) {
		addHandler(sourceClass, targetClass, provider);
	}

	/**
	 * Returns the first {@link ITransformationProvider} which can connect the
	 * given elements (or <code>null</code>n if no such connector exists).
	 */
	private ITransformationProvider findWorkingTransformationProvider(Object source,
			Class<?> targetClass, ITransformationContext context) {
		List<ITransformationProvider> list = getRegisteredHandlers(source.getClass(), targetClass);
		if(list == null) {
			return null;
		}
		for(ITransformationProvider provider : list) {
			if(provider.canTransform(source, context)) {
				return provider;
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public boolean
			canTransform(Object source, Class<?> targetClass, ITransformationContext context) {
		return findWorkingTransformationProvider(source, targetClass, context) != null;
	}

	/** {@inheritDoc} */
	@Override
	public Object transform(Object source, Class<?> targetClass, ITransformationContext context)
			throws TransformationFailedException {
		return findWorkingTransformationProvider(source, targetClass, context).transform(source,
				context);
	}

	/** {@inheritDoc} */
	@Override
	public List<TransformationProviderChain> getTransformationProviderChain(Class<?> sourceClass,
			Class<?> targetClass, ITransformationContext context) {
		List<TransformationProviderChain> results = new LinkedList<TransformationProviderChain>();
		List<ITransformationProvider> currentChain = new LinkedList<ITransformationProvider>();
		Set<Class<?>> seenTargets = new HashSet<Class<?>>();
		seenTargets.add(sourceClass);
		findTransformationChains(sourceClass, targetClass, results, currentChain, seenTargets,
				context);
		return results;
	}

	/**
	 * Recursive implementation of transformation provider search. This method
	 * is shielded against infinite loops within the transformation provider
	 * graph.
	 */
	private void findTransformationChains(Class<?> sourceClass, Class<?> targetClass,
			List<TransformationProviderChain> results, List<ITransformationProvider> currentChain,
			Set<Class<?>> seenTargets, ITransformationContext context) {
		if(!currentChain.isEmpty() &&
				targetClass.isAssignableFrom(currentChain.get(currentChain.size() - 1)
						.getTargetClass())) {
			// found a complete chain => add to result and backtrack
			results.add(new TransformationProviderChain(currentChain));
			return;
		}
		if(seenTargets.contains(targetClass)) {
			// infinite loop protection
			return;
		}
		Map<Class<?>, List<ITransformationProvider>> providerMap =
				ReflectionUtils.performNearestClassLookup(sourceClass, handlersBySource);
		if(providerMap == null || providerMap.isEmpty()) {
			// dead end reached
			return;
		}
		for(Class<?> key : providerMap.keySet()) {
			List<ITransformationProvider> providerList = providerMap.get(key);
			for(ITransformationProvider provider : providerList) {
				if(provider.canHandleChainTransformation(sourceClass, context)) {
					// recursion starts here, add provider and loop protection
					currentChain.add(provider);
					seenTargets.add(provider.getTargetClass());
					findTransformationChains(provider.getTargetClass(), targetClass, results,
							currentChain, seenTargets, context);
					// backtrack
					currentChain.remove(provider);
					seenTargets.remove(provider.getTargetClass());
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Transformation Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new TransformationServiceIntrospectionDetailsItem(handlersBySource);
	}
}
