/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2016 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.util.Collections.emptyMap;
import static org.conqat.lib.commons.collections.CollectionUtils.asUnmodifiable;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EventObject;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.ITutorialProvider;
import org.fortiss.tooling.kernel.extension.ITutorialWhitelistProvider;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.extension.data.Prototype;
import org.fortiss.tooling.kernel.extension.data.TutorialAtomicStep;
import org.fortiss.tooling.kernel.extension.data.TutorialCompositeStep;
import org.fortiss.tooling.kernel.extension.data.TutorialStepBase;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.ITutorialService;
import org.fortiss.tooling.kernel.service.listener.ITutorialServiceListener;

/**
 * This class implements the {@link ITutorialService} interface.
 * 
 * @author hoelzl
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating YELLOW Hash: 3461DE607ABE30B3414A10D064153A7A
 */
public final class TutorialService implements ITutorialService, CommandStackListener,
		IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final TutorialService INSTANCE = new TutorialService();

	/** Returns singleton instance of the service. */
	public static TutorialService getInstance() {
		return INSTANCE;
	}

	/** The collection of all defined tutorials. */
	private Map<String, Map<String, Class<? extends ITutorialProvider>>> allTutorials =
			new HashMap<>();

	/** The collection of listeners. */
	private Collection<ITutorialServiceListener> listeners =
			new LinkedList<ITutorialServiceListener>();

	/** The list of atomic steps. */
	private List<TutorialAtomicStep> atomicStepList = new ArrayList<TutorialAtomicStep>();

	/** The last element of the atomic step list, i.e., the final step. */
	private TutorialAtomicStep finalAtomicStep;

	/** The active tutorial. */
	private ITutorialProvider activeTutorial = null;

	/** The active tutorial step. */
	private TutorialAtomicStep activeStep = null;

	/** The active top-level element. */
	private ITopLevelElement activeTopElement = null;

	/** The persistency service. */
	private IPersistencyService persistencyService;

	/** The command stack service. */
	private ICommandStackService commandStackService;

	/** Initializes the service. */
	public void initializeService() {
		// nothing to do here
	}

	/** Starts the service. */
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Tutorial Service";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service manages tutorials, which can be registered via it." +
				"\nThis service can be asked to start and to stop tutorials. It also provides information " +
				"\nabout the currently active tutorial and the current tutorial step." +
				"\nIt notifies interested listeners about progress and changes to the current step." +
				"\nThis service provides access to prototypes constrained by the current step." +
				"\nThe other kernel services interact with the tutorial service when a tutorial is active" +
				"\nin order to provide the selective look-and-feel of the tutorial mode.";
	}

	/** {@inheritDoc} */
	@Override
	public boolean isTutorialActive() {
		return activeTutorial != null;
	}

	/** {@inheritDoc} */
	@Override
	public ITutorialProvider getActiveTutorial() {
		return activeTutorial;
	}

	/** {@inheritDoc} */
	@Override
	public TutorialStepBase getActiveTutorialStep() {
		return activeStep;
	}

	/** {@inheritDoc} */
	@Override
	public Map<String, Class<? extends ITutorialProvider>> getAllTutorialProviders(String category) {
		if(!allTutorials.containsKey(category)) {
			return emptyMap();
		}
		return asUnmodifiable(allTutorials.get(category));
	}

	/** {@inheritDoc} */
	@Override
	public Collection<String> getProviderCategories() {
		return asUnmodifiable(allTutorials.keySet());
	}

	/** {@inheritDoc} */
	@Override
	public void registerTutorialProvider(Class<? extends ITutorialProvider> provider,
			String displayName, String category) {
		if(category == null) {
			category = "";
		}
		Map<String, Class<? extends ITutorialProvider>> categoryMap = allTutorials.get(category);
		if(categoryMap == null) {
			categoryMap = new HashMap<String, Class<? extends ITutorialProvider>>();
			allTutorials.put(category, categoryMap);
		}
		categoryMap.put(displayName, provider);
	}

	/** {@inheritDoc} */
	@Override
	public void startTutorial(Class<? extends ITutorialProvider> provider) {
		persistencyService = IPersistencyService.getInstance();
		commandStackService = ICommandStackService.getInstance();
		try {
			activeTutorial = provider.newInstance();
			List<TutorialStepBase> rootSteps = activeTutorial.getDefinition().getSteps();
			for(TutorialStepBase step : rootSteps) {
				collectAtomicSteps(step);
			}
			// there is always at least one atomic step, therefore get(0) is fine
			activeStep = atomicStepList.get(0);
			// there is always at least one atomic step, therefore get(last) is fine
			finalAtomicStep = atomicStepList.get(atomicStepList.size() - 1);
			EObject root = activeTutorial.getRootElement();
			activeTopElement = persistencyService.addDummyEObjectAsTopLevelElement(root);
			commandStackService.addCommandStackListener(activeTopElement, this);
			fireTutorialStarted();
		} catch(InstantiationException e) {
			error(ToolingKernelActivator.getDefault(), e.getMessage(), e);
		} catch(IllegalAccessException e) {
			error(ToolingKernelActivator.getDefault(), e.getMessage(), e);
		}
	}

	/** Collects all atomic steps. */
	private void collectAtomicSteps(TutorialStepBase step) {
		if(step instanceof TutorialAtomicStep) {
			atomicStepList.add((TutorialAtomicStep)step);
		} else if(step instanceof TutorialCompositeStep) {
			TutorialCompositeStep tsc = (TutorialCompositeStep)step;
			for(TutorialStepBase s : tsc.getSteps()) {
				collectAtomicSteps(s);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void stopTutorial() {
		fireTutorialStopped();
		commandStackService.removeCommandStackListener(activeTopElement, this);
		commandStackService = null;
		persistencyService.removeDummyTopLevelElement(activeTutorial.getRootElement());
		persistencyService = null;
		activeTopElement = null;
		atomicStepList.clear();
		activeStep = null;
		activeTutorial = null;
	}

	/** {@inheritDoc} */
	@Override
	public void addTutorialServiceListener(ITutorialServiceListener l) {
		if(!listeners.contains(l)) {
			listeners.add(l);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeTutorialServiceListener(ITutorialServiceListener l) {
		listeners.remove(l);
	}

	/** {@inheritDoc} */
	@Override
	public boolean prototypeActive(Prototype prototype) {
		TutorialStepBase curStep = activeStep;
		while(curStep != null) {
			if(curStep instanceof ITutorialWhitelistProvider) {
				ITutorialWhitelistProvider p = (ITutorialWhitelistProvider)curStep;
				if(p.prototypeActive(prototype)) {
					return true;
				}
			}
			curStep = curStep.getParent();
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isCompletedStep(TutorialStepBase step) {
		if(step instanceof TutorialCompositeStep) {
			TutorialCompositeStep cmp = (TutorialCompositeStep)step;
			TutorialStepBase inner = cmp.getSteps().get(cmp.getSteps().size() - 1);
			return isCompletedStep(inner);
		}
		// step is of type TutorialStepAtomic
		if(step == activeStep) {
			return activeStep.getCompletionChecker().test(activeTutorial.getRootElement());
		}
		if(isEarlierThanActiveStep(step)) {
			return true;
		}
		return false;
	}

	/** Returns whether the given step is an earlier step than the current one. */
	private boolean isEarlierThanActiveStep(TutorialStepBase step) {
		if(step instanceof TutorialCompositeStep) {
			TutorialCompositeStep cmp = (TutorialCompositeStep)step;
			TutorialStepBase inner = cmp.getSteps().get(cmp.getSteps().size() - 1);
			return isEarlierThanActiveStep(inner);
		}
		// step is of type TutorialStepAtomic
		int current = atomicStepList.indexOf(activeStep);
		int test = atomicStepList.indexOf(step);
		return test < current;
	}

	/** {@inheritDoc} */
	@Override
	public void commandStackChanged(EventObject event) {
		TutorialAtomicStep old = activeStep;
		activeStep = finalAtomicStep;
		for(TutorialAtomicStep step : atomicStepList) {
			if(!step.getCompletionChecker().test(activeTutorial.getRootElement())) {
				activeStep = step;
				break;
			}
		}
		if(activeStep != old || activeStep == finalAtomicStep) {
			fireTutorialStepChanged();
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<TutorialAtomicStep> getAtomicTutorialSteps() {
		return asUnmodifiable(atomicStepList);
	}

	/** Fires {@link ITutorialServiceListener#tutorialStepChanged(TutorialAtomicStep)} event. */
	private void fireTutorialStepChanged() {
		for(ITutorialServiceListener l : listeners) {
			l.tutorialStepChanged(activeStep);
		}
	}

	/** Fires {@link ITutorialServiceListener#tutorialStarted(ITutorialProvider)} event. */
	private void fireTutorialStarted() {
		for(ITutorialServiceListener l : listeners) {
			l.tutorialStarted(activeTutorial);
		}
	}

	/** Fires {@link ITutorialServiceListener#tutorialStopped(ITutorialProvider)} event. */
	private void fireTutorialStopped() {
		for(ITutorialServiceListener l : listeners) {
			l.tutorialStopped(activeTutorial);
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
